//////////////////////////////////////////////////////////////////////////
// Gp32disk: GP32 Disk and File API emulation library
// Copyright (C) 2008-2009, hardyx
// This library can be used freely under the GNU GPL license.
//////////////////////////////////////////////////////////////////////////

#include "gp32emu.h"
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <dirent.h>

// prototipos
void GpVideoDone();

// variables privadas
static bool m_diskInit = false;
static char m_pathapp[GP_MAX_PATH];

//////////////////////////////////////////////////////////
// funciones de disco

// inicializa el sistema de ficheros
ERR_CODE GpFatInit()
{
	if(m_diskInit) {
		printf("GpFat ya iniciado\n");
		return ERR_INVALID_FAT;
	}

  // guardar el path del programa
  getcwd(m_pathapp, sizeof(m_pathapp));
  strcat(m_pathapp,SDIR_SEP);
	m_diskInit = true;
  printf("GpFatInit\n");
	return SM_OK;
}

// indica si est inicializado
ERR_CODE GpFatInitialized()
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	return SM_OK;
}

// path de la aplicacin actual
char *GpAppPathGet(int *n_len)
{
  if(n_len) // opcional
    *n_len = strlen(m_pathapp);
  return m_pathapp;
}

// terminar app
void GpAppExit(void)
{
  printf("GpAppExit\n");
  GpVideoDone();
#ifdef GP2X
	chdir("/usr/gp2x");
	execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
  printf("Despus de gp2xmenu\n");
#endif
	exit(0);
}

// devuelve el path sin carpetas de la gp32
static const char *SkipGpRoot(const char *gppath)
{
	int len = strlen(gppath);
	if(len>=4 && !strncmp(gppath,"gp:\\",4))
		return gppath+4;
	return gppath;
}

// crea el directorio dado, los flags no se usan
ERR_CODE GpDirCreate(const char *dir_name, ulong dcreate_mode/*=0*/)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!dir_name)
		return ERR_INVALID_PARAM;

	const char *dname = SkipGpRoot(dir_name);
#ifdef WIN32
	if(mkdir(dname) != 0)
#else
	if(mkdir(dname,0) != 0)
#endif
		return ERR_INVALID_PARAM;
	return SM_OK;
}

// borra el directorio segn las opciones
ERR_CODE GpDirRemove(const char *dir_name, ulong ddel_mode)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!dir_name)
		return ERR_INVALID_PARAM;
	if(ddel_mode == ALWAYS_DELETE) // no soportado
		return ERR_DIR_NOT_EMPTY;

	const char *dname = SkipGpRoot(dir_name);
	if(rmdir(dname) != 0)
		return ERR_INVALID_PARAM;
	return SM_OK;
}

// devuelve true si el directorio existe
bool GpDirExists(const char *dir_name)
{
	if(!dir_name || !*dir_name)
		return false;

	const char *dname = SkipGpRoot(dir_name);
	if(access(dname,0) != 0)
		return false;
	return true;
}

// da el nmero de elementos de un diretorio
ERR_CODE GpDirEnumNum(const char *dir_name, ulong *p_num)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!dir_name || !p_num)
		return ERR_INVALID_PARAM;

	const char *dname = SkipGpRoot(dir_name);
	bool inroot = false;
#ifdef GP2X
  // no permitir listar en la nand
	//inroot = !strcmp(dir_name,"/mnt/") || !strcmp(dir_name,"/tmp/mnt/");
#else
  // pruebas en windows
	//inroot = !strcmp(dir_name,"c:\\mios\\");
#endif

  DIR *dir = opendir(dname);
	ulong num = 0;
  if(dir) {
    while (readdir(dir)) num++;
    closedir(dir);
  }

  if(inroot) num--;
	*p_num = num;
	return SM_OK;
}

// da una lista con los elementos de un directorio
// desde la posicin entry_start da entry_count elementos
ERR_CODE GpDirEnumList(const char *dir_name, ulong entry_start, ulong entry_count,
											 GPDIRENTRY *list, ulong *read_count)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!dir_name || entry_count == 0)
		return ERR_INVALID_PARAM;

	const char *dname = SkipGpRoot(dir_name);
	bool inroot = false;
#ifdef GP2X
  // no permitir listar en la nand
	//inroot = !strcmp(dir_name,"/mnt/") || !strcmp(dir_name,"/tmp/mnt/");
#else
  // pruebas en windows
	//inroot = !strcmp(dir_name,"c:\\mios\\");
#endif

	ulong entry_end = entry_start+entry_count-1;
	ulong index=0, leidos=0;
  DIR *dir = opendir(dname);

  if(dir) {
		struct dirent *de = readdir(dir);
    while (de) {
      if(inroot && !strcmp(de->d_name,"..")) { // ignorar
        de = readdir(dir); // leer sig
        continue;
      }
			if(index>=entry_start && index<entry_end) {
				sprintf(list->name,"%.*s",GP_MAX_PATH,de->d_name);
				leidos++;
				list++; // siguiente elem
			}
			index++;
			de = readdir(dir);
		}
    closedir(dir);
  }

	*read_count = leidos;
	return SM_OK;
}

// da los atributos de un fichero
ERR_CODE GpFileAttr(const char *name, GPFILEATTR *attr)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!name || !attr)
		return ERR_INVALID_PARAM;

	const char *fname = SkipGpRoot(name);
	memset(attr,0,sizeof(GPFILEATTR));

  // implementacin mejor
  struct stat st;
  if(stat(fname,&st) == -1)
    return ERR_FILE_NOT_EXIST;

	if(S_ISREG(st.st_mode))
     attr->attr = 0x20; // archivo
  else if(S_ISDIR(st.st_mode))
     attr->attr = 0x10; // directorio
  else
     attr->attr = 0x06; // sys+hidden

  // tamao del archivo
  attr->size = (uword)st.st_size;

	// obtener los campos de la fecha
	struct tm *date = gmtime(&st.st_mtime);
	attr->time.year = (uword)date->tm_year + 1900;
	attr->time.month = (ubyte)date->tm_mon;
	attr->time.day = (ubyte)date->tm_mday;
	attr->time.hour = (ubyte)date->tm_hour;
	attr->time.min = (ubyte)date->tm_min;
	attr->time.sec = (ubyte)date->tm_sec;
	return SM_OK;
}

//////////////////////////////////////////////////////////
// funciones de ficheros

// abre el fichero en el modo indicado y da su handle
ERR_CODE GpFileOpen(const char *file_name, ulong fopen_mode, F_HANDLE *handle)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!handle)
		return ERR_INVALID_PARAM;

	const char *fname = SkipGpRoot(file_name);
	int flags = 0;
#ifdef WIN32
  flags |= O_BINARY;
#endif
	if(fopen_mode == OPEN_R)
		flags |= O_RDONLY;
	else if(fopen_mode == OPEN_W)
		flags |= O_WRONLY;
	else if(fopen_mode == OPEN_R+OPEN_W)
		flags |= O_RDWR;

	*handle = (long)open(fname,flags);
	if(*handle == -1)
		return ERR_FILE_NOT_OPENED;
	return SM_OK;
}

// crea el fichero en el modo indicado y da su handle
ERR_CODE GpFileCreate(const char *file_name, ulong fcreate_mode, F_HANDLE *handle)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!file_name || !handle)
		return ERR_INVALID_PARAM;

	const char *fname = SkipGpRoot(file_name);
	int flags = O_CREAT;
#ifdef WIN32
  flags |= O_BINARY;
#endif
	if(fcreate_mode == NOT_IF_EXIST)
		flags |= O_EXCL | O_RDWR;
	else if(fcreate_mode == ALWAYS_CREATE)
		flags |= O_TRUNC | O_RDWR;
	// permisos del fichero creado
	int pmode = S_IREAD | S_IWRITE;

	*handle = (long)open(fname,flags,pmode);
	if(*handle == -1)
		return ERR_INTERNAL;
	return SM_OK;
}

// cierra el fichero
ERR_CODE GpFileClose(F_HANDLE h_file)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!h_file)
		return ERR_INVALID_HANDLE;

	if(close(h_file) == -1)
		return ERR_FILE_NOT_OPENED;
	return SM_OK;
}

// lee un bloque de un fichero
ERR_CODE GpFileRead(F_HANDLE h_file, void *buf, ulong buf_size, ulong *read_count)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!h_file)
		return ERR_INVALID_HANDLE;
	if(!read_count || !buf)
		return ERR_INVALID_PARAM;

	long bytes = read(h_file,buf,buf_size);
	if(bytes == -1)
		return ERR_INTERNAL;
	*read_count = (ulong)bytes;
	return SM_OK;
}

// escribe un bloque en el fichero
ERR_CODE GpFileWrite(F_HANDLE h_file, const void *buf, ulong count)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!h_file)
		return ERR_INVALID_HANDLE;
	if(!buf)
		return ERR_INVALID_PARAM;

	if(write(h_file,buf,count) == -1)
		return ERR_INTERNAL;
	return SM_OK;
}

// da el tamao de un fichero
ERR_CODE GpFileGetSize(const char *file_name, ulong *size)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!file_name || !size)
		return ERR_INVALID_PARAM;

	const char *fname = SkipGpRoot(file_name);

  // unix no tiene _filelength ni _tell
  FILE *file = fopen(fname,"rb");
  if(file == NULL)
		return ERR_FILE_NOT_EXIST;

  long oldpos = ftell(file);
  fseek(file,0,SEEK_END);
  long lsize = ftell(file);
  fseek(file,oldpos,SEEK_SET);
  fclose(file);

	if(lsize == -1)
		return ERR_INTERNAL;
	*size = (ulong)lsize;
	return SM_OK;
}

// se posiciona en un offset dado del fichero
ERR_CODE GpFileSeek(F_HANDLE h_file, ulong seek_mode, ulong offset, long *old_offset)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!h_file)
		return ERR_INVALID_HANDLE;
	if(!old_offset || seek_mode>2)
		return ERR_INVALID_PARAM;

	const int modes[3] = {SEEK_CUR,SEEK_SET,SEEK_END};
	*old_offset = lseek(h_file,offset,modes[seek_mode]);
	if(*old_offset == -1)
		return ERR_NOT_FOUND;
	return SM_OK;
}

// borra el fichero dado
ERR_CODE GpFileRemove(const char *file_name)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!file_name)
		return ERR_INVALID_PARAM;

	const char *fname = SkipGpRoot(file_name);
	if(remove(fname) == -1)
		return ERR_NOT_PERMITTED;
	return SM_OK;
}

// cambia el nombre al fichero por otro
ERR_CODE GpFileRename(const char *old_path, const char *new_path)
{
	if(!m_diskInit)
		return ERR_NOT_INITIALIZED;
	if(!old_path || new_path)
		return ERR_INVALID_PARAM;

	const char *fname = SkipGpRoot(old_path);
	const char *fname2 = SkipGpRoot(new_path);
	if(rename(fname,fname2) != 0)
		return ERR_NOT_PERMITTED;
	return SM_OK;
}

