//////////////////////////////////////////////////////////////////////////
// Gp32video: GP32 Video API emulation library
// Copyright (C) 2008-2009, hardyx
// This library can be used freely under the GNU GPL license.
//////////////////////////////////////////////////////////////////////////

#include "gp32emu.h"
#include <stdlib.h>
#include <string.h>
#include <SDL/SDL.h>

// conversin antigua
//#define CONV_PIXFORMAT4

// constantes usadas
#define WIDTH  320
#define HEIGHT 240

#ifdef WIN32
#define FULLSCR 0
#else
#define FULLSCR 1
#endif

#ifdef CONV_PIXFORMAT4
// pixelformat de la gp32 (5551)
#define _gp_rmask 0xF800
#define _gp_gmask 0x07C0
#define _gp_bmask 0x003E
#define _gp_rshift 11
#define _gp_gshift  6
#define _gp_bshift  1
#endif

//////////////////////////////////////////////////////////
// framebuffer de la gp32 (est girado)
typedef short TFrameBuf[HEIGHT*WIDTH];

//////////////////////////////////////////////////////////
// funciones y variables privadas

// indica si la sdl est inicializada
bool m_sdlInit = false;

// objeto joystick, botones gp2x
SDL_Joystick *m_joystick = 0;

// superficie pantalla
static SDL_Surface *m_screen = 0;

// offsets de filas para no multiplicar
static int m_gpcolums[WIDTH];

// framebuffers de la gp32
static TFrameBuf m_frame[2];
static int m_factive = 0;

void GpVideoDone();

//////////////////////////////////////////////////////////
// funciones generales y de teclado

void GpVideoInit()
{
	if(m_sdlInit) {
		printf("GpVideo ya iniciado\n");
		exit(1);
	}
	// esto por si hay un fallo grave
	atexit(GpVideoDone);
	m_factive = 0;

	// inicializar la librera SDL
	if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK) < 0) {
		printf("Error al iniciar SDL: %s\n", SDL_GetError());
		exit(1);
	}

	// inicia el retardo del teclado o pad
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY,SDL_DEFAULT_REPEAT_INTERVAL);

	// precalcular offsets de columnas
	for(int i=0; i<WIDTH; i++)
		m_gpcolums[i] = i*HEIGHT;

	m_sdlInit = true;
	printf("GpVideo init ok!\n");
}

void GpVideoDone()
{
	if(m_sdlInit) {
		if(m_joystick) {
			// liberar el joystick
			if(SDL_JoystickOpened(0))
				SDL_JoystickClose(0);
			m_joystick = NULL;
		}

		SDL_Quit();
		m_sdlInit = false;
	  printf("GpVideo done!\n");
	}
}

//////////////////////////////////////////////////////////
// funciones de superficies y LCD

// inicia el modo grfico, da el nmero de buffers
int GpGraphicModeSet(int gd_bpp, int *gp_pal/*=0*/)
{
  printf("GpGraphicModeSet\n");
	if(!m_sdlInit) {
		printf("GpVideo no iniciado\n");
		return 0;
	}
	if(gd_bpp!=16 || gp_pal) {
		printf("Slo soportado 16 bits de color\n");
		return 0;
	}

	// iniciar el modo grfico
	// el sdk oficial no soporta superficies hardware
	Uint32 vidflags;
	if (FULLSCR)
		vidflags = SDL_FULLSCREEN | SDL_SWSURFACE;
	else
		vidflags = SDL_SWSURFACE;

	m_screen = SDL_SetVideoMode(WIDTH,HEIGHT,16,vidflags);
	if(!m_screen) {
		printf("Error en SetVideoMode: %s\n", SDL_GetError());
		return 0;
	}

	// desactivar cursor en fullscreen
	SDL_ShowCursor(FULLSCR? SDL_DISABLE: SDL_ENABLE);

	// abrir el joystick (botones)
	if(SDL_NumJoysticks() > 0) {
		m_joystick = SDL_JoystickOpen(0);
		if(!m_joystick)
			printf("Error al abrir el joystick\n");
		SDL_JoystickEventState(SDL_ENABLE);
	}

	return 2;
}

// da la superficie hardware con dicho ndice
int GpLcdSurfaceGet(GPDRAWSURFACE *ptgpds, int idx)
{
	if(!m_sdlInit || !ptgpds)
		return GPC_DRAW_ERR_PARAM;
	if(idx<0 || idx>1)
		return GPC_DRAW_ERR_PARAM;

	// guardamos ndice de la superficie
	ptgpds->i_buffer = idx;
	// puntero a los datos de la imagen
	ptgpds->ptbuffer = &m_frame[idx][0];
	ptgpds->ox = ptgpds->oy = 0;
	ptgpds->buf_w = WIDTH;
	ptgpds->buf_h = HEIGHT;
	ptgpds->bpp = 16;
	return GPC_DRAW_OK;
}

// registra el buffer para transferir al lcd
void GpSurfaceSet(GPDRAWSURFACE *ptgpds)
{
	if(!m_sdlInit || !ptgpds || !m_screen)
		return;

	// volcar el framebuffer a la pantalla
	SDL_LockSurface(m_screen);

	// copiar el framebuffer a la pantalla
	short *screen = (short*)m_screen->pixels;
	const short *ptbuffer = ptgpds->ptbuffer;
	int pitch_dif = WIDTH - m_screen->pitch/2;

	// se usa gpcolumns para no multiplicar
	for(register int y=0; y<HEIGHT; y++) {
		for(register int x=0; x<WIDTH; x++)
			*screen++ = ptbuffer[m_gpcolums[x] + (HEIGHT-1-y)];
		screen += pitch_dif;
	}

#ifdef CONV_PIXFORMAT4
  // funciona!!
	// convertir pixelformat gp32 a gp2x (551 a 565)
	ushort *uscreen = (ushort*)m_screen->pixels;
	int pitch = m_screen->pitch/2;
	for(register int i=0; i<pitch*HEIGHT; i++) {
	  ushort rg = uscreen[i] & 0xFFC0;
	  ushort b = (uscreen[i] & 0x003E) >> 1;
    uscreen[i] = rg | b;
	}
#endif

	// actualiza la pantalla entera
	SDL_UnlockSurface(m_screen);
	SDL_UpdateRect(m_screen,0,0,0,0);

	// slo asignar un buffer vlido
	int ibuf = ptgpds->i_buffer;
	if(ibuf==0 || ibuf==1)
		m_factive = ibuf;
}

// reeemplaza el buffer del lcd con uno nuevo
void GpSurfaceFlip(GPDRAWSURFACE *ptgpds)
{
  //printf("-GpSurfaceFlip\n");
	if(!m_sdlInit || !ptgpds)
		return;
	GpSurfaceSet(ptgpds);

	// cambiar el framebuffer
	m_factive = m_factive? 0: 1;
}

//////////////////////////////////////////////////////////
// funciones de dibujo

__inline short *GetPixel(GPDRAWSURFACE *ptgpds,int x,int y)
{
	if(ptgpds->ptbuffer)
		return &(ptgpds->ptbuffer[m_gpcolums[x] + y]);
	return 0;
}

// dibuja un punto en x,y
void GpPointSet16(GPDRAWSURFACE *ptgpds,int x,int y,int color)
{
	if(!m_sdlInit || !ptgpds)
		return;
	if(x>=WIDTH || y>=HEIGHT)
		return;
	*GetPixel(ptgpds,x,y) = (short)color;
}

// dibuja una lnea vertical de x,ys a x,ye
void _LineDrawVert16(GPDRAWSURFACE *ptgpds,int x,int yS,int yE,int color)
{
	if(!m_sdlInit || !ptgpds)
		return;
	if(x>=WIDTH || yS>=HEIGHT || yE>=HEIGHT)
		return;
	if(yS <= yE) {
		// pintar pixels ys a ye (bloque)
		short *buff = GetPixel(ptgpds,x,yS);
		memset(buff,(short)color,yE-yS+1);
	}
}

// dibuja una lnea horizontal de xs,y a xe,y
void _LineDrawHor16(GPDRAWSURFACE *ptgpds,int xS,int xE,int y,int color)
{
	if(!m_sdlInit || !ptgpds)
		return;
	if(xS>=WIDTH || xE>=WIDTH || xS>xE || y>=HEIGHT)
		return;

	// pintar pixels xs a xe
	short *buff = GetPixel(ptgpds,xS,y);
	for(int i=xS; i<=xE; i++)
		buff[i] = color;
}

// dibuja una lnea de xs,ys a xe,ye (horiz o vert)
void GpLineDraw16(GPDRAWSURFACE *ptgpds,int xS,int yS,int xE,int yE,int color)
{
	if(!m_sdlInit || !ptgpds)
		return;
	if(xS>=WIDTH || yS>=HEIGHT || xE>=WIDTH || yE>=HEIGHT)
		return;
	if(xS>xE || yS>yE)
		return;
	if(xS == xE)
		_LineDrawVert16(ptgpds,xS,yS,yE,color);
	else if(yS == yE)
		_LineDrawHor16(ptgpds,xS,xE,yS,color);
	else {
		printf("GpLineDraw: linea diagonal\n");
		return; // no implementado
	}
}

// dibuja un rectngulo relleno en dx,dy,dx+w,dy+h
int GpRectFill16(GPDRAWTAG *gptag,GPDRAWSURFACE *ptgpds,int dx,int dy,int width,int height,int color)
{
	if(!m_sdlInit || !ptgpds)
		return -1;
	if(dx>=WIDTH || dy>=HEIGHT || dx+width>WIDTH || dy+height>HEIGHT)
		return -1;
	if(width<0 || height<0)
		return -1;

	// pintarlo con lineas verticales
	for(int i=dx; i<dx+width; i++) {
		short *buff = GetPixel(ptgpds,i,dy);
		memset(buff,(short)color,height);
	}
	return GPC_DRAW_OK;
}

// dibuja un rectngulo en l,t,r,b
void GpRectDraw16(GPDRAWSURFACE *ptgpds,int left,int top,int right,int bottom,int color)
{
	if(!m_sdlInit || !ptgpds)
		return;
	if(left>=WIDTH || top>=HEIGHT || right>=WIDTH || bottom>=HEIGHT)
		return;
	if(left>right || top<bottom) // y al revs
		return;

	// usar lneas horizonts y verticales
	_LineDrawHor16(ptgpds,left,right,top,color);
	_LineDrawHor16(ptgpds,left,right,bottom,color);
	_LineDrawVert16(ptgpds,left,top,bottom,color);
	_LineDrawVert16(ptgpds,right,top,bottom,color);
}

// no implementar
int GpPaletteEntryChange(int pal_offset, int entry_num, GP_PALETTEENTRY *pal_entry, int flag)
{
	return GPC_DRAW_OK;
}
